//**************************************************************************
//*                     This file is part of the                           *
//*                 Mpxplay/MMC - multimedia player.                       *
//*                  The source code of Mpxplay is                         *
//*        (C) copyright 1998-2019 by PDSoft (Attila Padar)                *
//*                http://mpxplay.sourceforge.net                          *
//*                  email: mpxplay@freemail.hu                            *
//**************************************************************************
//*  This program is distributed in the hope that it will be useful,       *
//*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
//*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                  *
//*  Please contact with the author (with me) if you want to use           *
//*  or modify this source.                                                *
//**************************************************************************
//function: FFMPEG library handling

//#define MPXPLAY_USE_DEBUGF 1
#define MPXPLAY_DEBUG_ERROR stderr
#define MPXPLAY_DEBUG_WARNING stderr
#define MPXPLAY_DEBUG_OUTPUT NULL //stdout
#define MPXPLAY_DEBUG_PLAY   NULL // stdout
#define MPXPLAY_DEBUG_OPENCLOSE  NULL // stdout
#define MPXPLAY_DEBUG_OPENCLSDTL NULL //stdout
#define MPXPLAY_DEBUG_HEADER  NULL //stdout
#define MPXPLAY_DEBUG_STDUR NULL // stdout
#define MPXPLAY_DEBUG_DEMUX NULL // stdout
#define MPXPLAY_DEBUG_ADECODE stdout
#define MPXPLAY_DEBUG_QUEUEGET NULL // stdout
#define MPXPLAY_DEBUG_QUEUEGET_AVFRAME NULL // stdout
#define MPXPLAY_DEBUG_QUEUEPUT NULL // stdout
#define MPXPLAY_DEBUG_TIMESTAMP NULL// stdout
#define MPXPLAY_DEBUG_STRMPRIO NULL // stdout
#define MPXPLAY_DEBUG_STRMLIST NULL // stdout
#define MPXPLAY_DEBUG_PRGCHG NULL // stdout
#define MPXPLAY_DEBUG_STRMCHG NULL // stdout
#define MPXPLAY_DEBUG_SEEK stdout
#define MPXPLAY_DEBUG_CALLBACK NULL // stdout
#define MPXPLAY_DEBUG_TEMP NULL // stdout
#define MPXPLAY_DEBUG_DEMUXSUB stdout

#include "in_file.h"

#ifdef MPXPLAY_LINK_INFILE_FF_MPEG

#define __STDC_CONSTANT_MACROS

#include <libavcodec/avcodec.h>
#include <libavformat/avformat.h>
#include <libavutil/pixfmt.h>

#ifdef MPXPLAY_USE_DEBUGF
 #define INFFMPG_DEBUG_HEADER 1
#endif

#define INFFMP_USE_DEMUXER_AUINIT 1 // FIXME: probably this solution is not thread safe, but required in some cases (flac + mjpeg)
#define INFFMP_SUPPORT_DVB_AUDIO  1 //
#define INFFMP_USE_VIDEO_SYNC     1 // start program with video sync (decode audio to this)
#define INFFMPG_USE_MPXPLAY_IOCTX 1 // use Mpxplay's onw low level file handling (dvb/ftp/http/file)
// these options shall be used together
#define INFFMP_USE_DEMUXER_THREAD 1 // TODO: frp buffer is not thread safe (pop and video freeze at start)
#define INFFMP_USE_DEMUXER_SEEK   1 // TODO: test
#define INFFMP_USE_DEMUXER_BUFILL 1

//#define INFFMP_USE_DEMUXER_MUTEX  1 // if demux thread and seek are separated (DEMUXER_THREAD is enabled and DEMUXER_SEEK is disabled)
//#define INFFMPG_USE_VIDEOPTS_CORRECTON 1 // for testing only (use avframe->pts instead)

#define INFFMPG_IOCTX_SIZE_MIN 4096
#define INFFMPG_IOCTX_SIZE_MAX 32768
#define INFFMPG_INITIAL_PACKETBUF_SIZE 65536 // TODO: this is good up to 13mbit video at stream length detect
#define INFFMPG_MAX_OUTPUT_SAMPLENUM  131072 // TODO: WMA (32k) & APE (80k) needs such large buffer
#define INFFMPG_MUTEX_TIMEOUT 5000
#define INFFMPG_MUTEX_QUEUEGET_TIMEOUT 10
#define INFFMPG_MUTEXTIME_NOLOCK  0x7eeeeeee
#define INFFMPG_QUEUE_PACKETNUM_MIN  3   // for both audio and video packets
#define INFFMPG_QUEUE_PACKETNUM_FILL_MIN  16
#define INFFMPG_QUEUE_PACKETNUM_FILL_STEP  8
#define INFFMPG_QUEUE_PACKETNUM_FILL_MAX  64
#define INFFMPG_QUEUE_PACKETNUM_MAX 1024  // to limit number of packets

#define INFFMPG_STREAM_INDEX_VALID    0  // >= 0 is a valid stream index
#define INFFMPG_STREAM_INDEX_DISABLE -1  // don't change these values
#define INFFMPG_INVALID_STREAM_INDEX -2  //
#define INFFMPG_STREAM_INDEX_LOST    -3  // current stream has lost

#define INFFMPG_STREAM_INDEX_EXTERNALFILE_BEGIN   65536 // stream belongs to a separated (audio, subtitle) file (not an internal stream)
#define INFFMPG_STREAM_INDEX_EXTERNALFILE_NB_MAX  64    // max number of (audio or subtitle) files in the list

#define INFFMPG_INVALID_PTS -1LL
#define INFFMPG_TIMESTAMP_SHIFT  (AV_TIME_BASE) // this must be larger than audio_delay in in_ffmpgdec_decode_audio
#define INFFMPG_TIMESTAMP_MAX_AVDIFFERENCE (AV_TIME_BASE / 2)  // max delay between video frames and audio/video frames
#define INFFMPG_MAX_FRAME_MISMACTHES 25 // * nb_programs
#define INFFMPG_MAX_STREAMS 255
#define INFFMPG_PTS_CHECK_NUM_MAX 12
#define INFFMPG_STILL_PICTURE_FRAME_MAX 3  // progressive jpeg may contain multiply frames
#define INFFMPG_SEEKREQUEST_INVALID_POS -1

#define INFFMPG_FLAG_AUTODETECT       (1 << 0)  // autodetect fileformat
#define INFFMPG_FLAG_LOADHEADONLY     (1 << 1)  // autodetect fileformat
#define INFFMPG_FLAG_IS_STREAM        (1 << 2)  // stream (MPG,TS)
#define INFFMPG_FLAG_IS_LIVESTREAM    (1 << 3)  // live stream
#define INFFMPG_FLAG_IS_SEEKABLE      (1 << 4)  // not live stream
#define INFFMPG_FLAG_STREAM_PARSING   (1 << 5)  // not live stream
#define INFFMPG_FLAG_INITIAL_SEEK     (1 << 6)  // don't seek to bof and don't clear buffs at first seek
#define INFFMPG_FLAG_SEEK_CLEAR       (1 << 7)  // clear ffctx bufs after seeking
#define INFFMPG_FLAG_INDEPENDENTSTREAMSYNC (1<<8)// do not synchronize A/V streams, every streams use it's own timestamp informations (mixing live stream(s))
#define INFFMPG_FLAG_PLANARAUDIO      (1 << 9)  // audio channels are separated (not interlaced)
#define INFFMPG_FLAG_STILLPICTURE     (1 <<10)  // jpg, bmp, etc.
#define INFFMPG_FLAG_CALLBACKVIDEOCFG (1 <<15)  // we've initialized a callback function for video surface (called back from Qt)
#define INFFMPG_FLAG_DEMUX_THREAD     (1 <<16)  // at video files
#define INFFMPG_FLAG_DEMUX_BUFREWIND  (1 <<17)  // rewind buffer at program change
#define INFFMPG_FLAG_DEMUX_TERMINATE  (1 <<18)  //
#define INFFMPG_FLAG_TEXTTYPE_BITPOS  (24)
#define INFFMPG_FLAG_TEXTTYPE_MASK    (0x03) //MPXPLAY_TEXTCONV_TYPE_CHAR (0),  MPXPLAY_TEXTCONV_TYPE_UTF8 (1), MPXPLAY_TEXTCONV_TYPE_UTF16LE (2), MPXPLAY_TEXTCONV_TYPE_UTF16BE (3)

#define INFFMPG_FLAGS_DIRECTCOPY_TO_PACKETTYPE ((1 << INFFMPG_FLAG_TEXTTYPE_BITPOS) | (1 << (INFFMPG_FLAG_TEXTTYPE_BITPOS + 1))) // ffmpeg_external_files_info_s->control_flags to packet_flags

#ifndef AVSEEK_FLAG_PAUSE
#define AVSEEK_FLAG_PAUSE    256
#endif

#define INFFMPG_STREAMFLAG_CLEARBUF_DEMUX  (1 << 0) // stream change or seek postprocess, flag for demultiplexer
#define INFFMPG_STREAMFLAG_AVSYNC          (1 << 2) // a/v sync by video
#define INFFMPG_STREAMFLAG_AVRESYNC        (1 << 3) // a/v resync by video if video packet queue is full (play video with silent audio (stop audio decoding))

#define INFFMPG_ERROR_STREAM_NOTPARSED        (-1000)
#define INFFMPG_ERROR_STREAM_LANGCODE_MISSING (-1001)  // stream has no language code (for inffmpg_stream_language_cmp)

#define INFFMPG_LANGCODE_ISO6391_LEN  2 // ISO 639-1 language code len
#define INFFMPG_LANGCODE_ISO6392_LEN  3 // ISO 639-2 language code len

typedef int (ffmpeg_read_packet_t)(void *, uint8_t *, int);

typedef struct ffmpeg_external_filelist_elem_s{
 unsigned int stream_index;                          // index begins at INFFMPG_STREAM_INDEX_EXTERNALFILE
 struct ffmpeg_external_filelist_elem_s *next_external_file;  // next external file elem in the chain
 enum AVCodecID codec_id;                            // stream informations
 int  bitrate, channels, samplerate;                 //
 char format_name[8];                                // file extension (like SRT,SUB or MP3)
 char language_code[4];                              // language code (like HUN or ENG)
 char language_name[64];                             // language (long) name
 char external_filename[MAX_PATHNAMELEN];            // filename with full path
}ffmpeg_external_filelist_elem_s;

typedef struct ffmpeg_external_files_info_s{
 struct ffmpeg_external_filelist_elem_s *external_files_chain;
 int nb_external_files;
 struct ffmpeg_external_filelist_elem_s *external_file_selected;
 struct mpxplay_filehand_buffered_func_s *fbfs;
 void *fbds;
 AVIOContext *ioctx;
 unsigned char *iobuffer;
 AVFormatContext *external_file_avctx;
 unsigned int streamtype_index;
 unsigned int control_flags;                         // INFFMPG_FLAG_
 mpxp_int64_t avclocktime_us;                        // reference clock time for independent stream sync
 mpxp_int64_t avtimestamp_us;                        // initial avtimestamp for this stream
}ffmpeg_external_files_info_s;

typedef struct ffmpg_demuxer_data_s{
 unsigned int flags, outsample_size, streamtype_limit;
 unsigned int stream_flags[MPXPLAY_STREAMTYPEINDEX_NUM];
 int retcode_demux;
 int errorcount_demux;
 int program_number_current;
 int select_new_program_number;
 int select_new_stream_number[MPXPLAY_STREAMTYPEINDEX_NUM];
 int selected_stream_number[MPXPLAY_STREAMTYPEINDEX_NUM];
 int selected_ff_stream_index[MPXPLAY_STREAMTYPEINDEX_NUM];
 unsigned int nb_streams[MPXPLAY_STREAMTYPEINDEX_NUM];
 unsigned int nb_frame_mismatches[MPXPLAY_STREAMTYPEINDEX_NUM];
 struct mpxplay_filehand_buffered_func_s *fbfs;
 void *fbds;
 struct mpxplay_infile_info_s *miis;
 unsigned int primary_file_openmode; // openmode for primary file
 AVInputFormat *primary_file_demuxer;// demuxer for main / primary file
 AVFormatContext *fctx;              // ffmpeg file handler context for main / primary file
 AVBufferRef *hw_device_ctx;         // for hardware decoding (dxva)
 enum AVPixelFormat hw_pix_fmt;      // for hardware decoding (dxva)
 struct ffmpeg_external_files_info_s external_file_infos[MPXPLAY_STREAMTYPEINDEX_NUM];
 AVStream *selected_avstream[MPXPLAY_STREAMTYPEINDEX_NUM];
 AVCodecContext *codec_ctx[MPXPLAY_STREAMTYPEINDEX_NUM];
 AVCodec *avcodecs[MPXPLAY_STREAMTYPEINDEX_NUM];
 mpxp_int64_t stream_last_pts[MPXPLAY_STREAMTYPEINDEX_NUM];
 AVPacket *pkt;
 AVFrame *audio_frame;
 int audioframe_codec_id;             // saved AV_CODEC_ID_ (for console versions only)
 int audioframe_channel_layout;       // saved channel layout
 int audioframe_samples_per_channel;  // samples per channel in audio_frame
 int audioframe_copy_sampleblocksize; // copy blocksize in samples (if we spilt a big frame for better a/v sync)
 int audioframe_copy_samplepos;       // sample pos of copy in audio_frame
#ifdef INFFMP_USE_DEMUXER_MUTEX
 void *mutexhnd_demuxer;            // mutex handler for demuxing (seek vs read thread)
#endif
 void *mutexhnd_ffmpeg;             // mutex handler for decoding
 mpxp_int64_t avsync_timestamp, avsync_clocktime;  // timestamp of audio, in AV_TIME_BASE
 mpxp_int64_t avsync_videotimestamp;// timestamp of displayed first video frame
 mpxp_int64_t seek_request_pos;     // in AV_TIME_BASE or byte position of the seek
 unsigned int seek_request_flags;   // flags for av_seek_frame, from FFMPG_infile_fseek_av to in_ffmpgdec_demux_read_packets
 int seek_request_retcode;          // result of the seek handled in an other thread
 int audioctx_init_error_count;     // a delay if audio context is not initialized (and because of this the packet is not used)
 mpxplay_packetqueue_t packet_queues[MPXPLAY_STREAMTYPEINDEX_NUM];    // audio/video data (frame) packets
 mpxplay_packetqueue_t codecctx_queues[MPXPLAY_STREAMTYPEINDEX_NUM];  // audio/video codec contexts
}ffmpg_demuxer_data_s;

#ifdef __cplusplus
extern "C" {
#endif

// in_fmpeg.c
extern mpxp_bool_t infmpeg_is_fileformat_video_stream(struct ffmpg_demuxer_data_s *ffmpi);
extern mpxp_bool_t infmpeg_is_fileformat_real_stream(struct ffmpg_demuxer_data_s *ffmpi);
extern int  in_fmpeg_audio_codecid_to_waveid(struct ffmpg_demuxer_data_s *ffmpi,struct mpxplay_infile_info_s *miis, AVFrame *audio_frame);
extern void inffmpg_tags_update(struct mpxplay_infile_info_s *miis, void *fbds, AVDictionary *metadata, unsigned int cmd);
// ffmppars.c
extern int in_ffmppars_parse_stream_header(struct ffmpg_demuxer_data_s *ffmpi, unsigned int stream_count_min);
extern mpxp_int64_t in_ffmppars_get_stream_duration_pts(struct ffmpg_demuxer_data_s *ffmpi, unsigned int checknum);
// ffmpgdec.c
extern void in_ffmpgdec_avsync_assign(struct ffmpg_demuxer_data_s *ffmpi, mpxp_int64_t pts);
extern mpxp_int64_t in_ffmpgdec_get_avpacket_ts_timestamp(struct ffmpg_demuxer_data_s *ffmpi, unsigned int streamtype_index, mpxp_int64_t pkt_ts);
extern mpxp_int64_t in_ffmpgdec_get_avpacket_timestamp(struct ffmpg_demuxer_data_s *ffmpi, unsigned int streamtype_index, AVPacket *pkt);
extern mpxp_int64_t in_ffmpgdec_select_avframe_pts(AVFrame *avframe);
extern mpxp_int64_t in_ffmpgdec_get_avframe_timestamp(struct ffmpg_demuxer_data_s *ffmpi, unsigned int streamtype_index, AVFrame *avframe);
extern void in_ffmpgdec_hwdec_deinit(AVCodecContext *codec_ctx);
extern int  in_ffmpgdec_infile_callback(void *demuxer_data, unsigned int control, mpxp_ptrsize_t arg1, mpxp_ptrsize_t arg2, int timeout_ms);
// ffmpadec.c
extern int in_ffmpgdec_decode_audio(struct ffmpg_demuxer_data_s *ffmpi, struct mpxplay_infile_info_s *miis);
// ffmpdmux.c
extern void in_ffmpdmux_seek_process_all(struct ffmpg_demuxer_data_s *ffmpi);
extern void in_ffmpdmux_read_packets(struct ffmpg_demuxer_data_s *ffmpi);
// ffmpfile.c
extern int  in_ffmpfile_filelist_file_add(struct ffmpg_demuxer_data_s *ffmpi, char *filename, int streamtype_index);
extern int  in_ffmpfile_filelist_files_add_lfcbds(struct ffmpg_demuxer_data_s *ffmpi, struct mpxplay_loadfile_callbackdata_s *mcbs);
extern void in_ffmpfile_filelist_load_list_by_ext(struct ffmpeg_external_files_info_s *extfiles, char *video_filename, unsigned int streamtype_index);
extern void in_ffmpfile_filelist_close(struct ffmpeg_external_files_info_s *extfiles);
extern int  in_ffmpfile_filelist_search_by_language(struct ffmpeg_external_files_info_s *extfiles, char *lang_code);
extern int  in_ffmpfile_file_open_by_streamindex(struct ffmpg_demuxer_data_s *ffmpi, unsigned int stream_type, int stream_index, AVStream **selected_avstream, AVCodec **codec_av_new);
extern void in_ffmpfile_file_close(struct ffmpeg_external_files_info_s *extfiles);
extern int  in_ffmpfile_file_packet_read(struct ffmpeg_external_files_info_s *extfiles, AVPacket *pkt);
extern int  in_ffmpfile_file_seek(struct ffmpeg_external_files_info_s *extfiles, mpxp_int64_t seek_pos, unsigned int seek_flags);
// ffmplang.c
extern const char *mpxplay_ffmplang_search_in_iso639_1to2_convs(const char *langcode_iso639_1_to_search);
extern const char *mpxplay_ffmplang_getlangname_by_langcode(const char *lang_code);
extern int mpxplay_ffmplang_get_language_fields(const char ***lang_codes_ptr, const char ***lang_names_ptr);
// ffmpqueu.c
extern unsigned int mpxplay_ffmpgdec_queue_clear_to_elemctx(mpxplay_packetqueue_t *packet_queue, void *ctx);
extern void in_ffmpgdec_packet_queues_clear(struct ffmpg_demuxer_data_s *ffmpi, mpxplay_packetqueue_t *packet_queue);
extern void in_ffmpgdec_packet_queues_init(mpxplay_packetqueue_t *packet_queue);
extern void in_ffmpgdec_packet_queues_close(struct ffmpg_demuxer_data_s *ffmpi, mpxplay_packetqueue_t *packet_queue);
// ffmpstrm.c
extern mpxp_bool_t in_ffmpgdec_content_has_programs(struct ffmpg_demuxer_data_s *ffmpi);
extern enum AVMediaType in_ffmpstrm_streamtypeindex_to_avmediatype(unsigned int streamtype_index);
extern mpxp_bool_t in_ffmpstrm_skip_program_auto(struct ffmpg_demuxer_data_s *ffmpi, int prognum_current_id);
extern int in_ffmpstrm_streams_count(struct ffmpg_demuxer_data_s *ffmpi, int streamtype_index, int *select_stream_number, int *selected_ff_streamindex, AVCodec **codec_av_new);
extern mpxp_bool_t in_ffmpgdec_streams_initialize_all(struct ffmpg_demuxer_data_s *ffmpi, mpxp_bool_t initial);
extern void in_ffmpgdec_stream_select_all_default(struct ffmpg_demuxer_data_s *ffmpi);

#ifdef __cplusplus
}
#endif

#endif //MPXPLAY_LINK_INFILE_FF_MPEG
