//**************************************************************************
//*                     This file is part of the                           *
//*                MMC - Mpxplay Multimedia Commander                      *
//*                   The source code of MMC is                            *
//*        (C) copyright 1998-2019 by PDSoft (Attila Padar)                *
//*                http://mpxplay.sourceforge.net                          *
//*                  email: mpxplay@freemail.hu                            *
//**************************************************************************
//*  This program is distributed in the hope that it will be useful,       *
//*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
//*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                  *
//*  Please contact with the author (with me) if you want to use           *
//*  or modify this source.                                                *
//**************************************************************************
//function: ffmpeg video decoding

//#define MPXPLAY_USE_DEBUGF 1
#define DISPQT_DEBUG_ERROR stderr
#define DISPQT_DEBUGOUT_WARNING stderr
#define DISPQT_DEBUGOUT_VTHREAD  NULL // stdout
#define DISPQT_DEBUGOUT_VFCC stdout

#include "moc_video_qt.h"

#define INFFMPG_MUTEX_ONTIMER_TIMEOUT     10 // used on 10ms timer
#define INFFMPG_MUTEX_PACKETGET_TIMEOUT  100
#define INFFMPG_MUTEX_COMMAND_TIMEOUT   5000

#ifdef MPXPLAY_LINK_ORIGINAL_FFMPEG

extern "C" {
 extern unsigned long mpxplay_config_video_audiovisualization_type;
}

static void video_worker_videodecoder_context_process(struct ffmpegvideo_callback_list_s *cbelem, AVCodecContext **video_ctx, bool *resetvideo, bool *resetfilter, unsigned int *sleep_time, unsigned int *packet_flags)
{
	int streamtype_index = MPXPLAY_STREAMTYPEINDEX_VIDEO, flags = MPXPLAY_PACKETLISTFLAG_CLEARBUF;
	mpxplay_packetlist_t *videoctx_pktelem = NULL;

	if((cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_GET_CODECCTXINIT_FFMPEG, (mpxp_ptrsize_t)&videoctx_pktelem, streamtype_index, INFFMPG_MUTEX_PACKETGET_TIMEOUT) == MPXPLAY_ERROR_OK) && videoctx_pktelem)
	{
		*video_ctx = (AVCodecContext *)videoctx_pktelem->codec_ctx;
		//mpxplay_debugf(DISPQT_DEBUGOUT_VTHREAD, "video_decoder_worker_thread GET_CODECCTXINIT pd:%8.8X ctx:%8.8X", (unsigned int)cbelem->infile_passdata, (unsigned int)video_ctx);
		if(!cbelem->video_pktlist_elem || (videoctx_pktelem->flags & MPXPLAY_PACKETLISTFLAG_CLEARBUF))
		{
			if(videoctx_pktelem->flags & MPXPLAY_PACKETLISTFLAG_CLEARBUF)
			{
				funcbit_disable(videoctx_pktelem->flags, MPXPLAY_PACKETLISTFLAG_CLEARBUF);
				*resetvideo = *resetfilter = true;
			}
			if(videoctx_pktelem->flags & MPXPLAY_PACKETLISTFLAG_STILLPICT)
				funcbit_enable(*packet_flags, MPXPLAY_PACKETLISTFLAG_STILLPICT);
			if( (cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_SET_PACKETQUEUE_CLEAR, streamtype_index, flags, INFFMPG_MUTEX_COMMAND_TIMEOUT) == 1)
			 || (*resetvideo && cbelem->video_pktlist_elem && (cbelem->video_pktlist_elem->codec_ctx != *video_ctx)) // TODO: check (wrong video_pktlist_elem->codec_ctx can crash the program)
			){
				if(cbelem->video_pktlist_elem){
					mpxplay_ffmpgdec_queuelistelem_clear(cbelem->video_pktlist_elem);
					cbelem->video_pktlist_elem = NULL;
				}
			}
			if(!cbelem->video_pktlist_elem)
				cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_GET_PACKETLIST_ELEM, (mpxp_ptrsize_t)&cbelem->video_pktlist_elem, streamtype_index, INFFMPG_MUTEX_PACKETGET_TIMEOUT);
			if(cbelem->video_pktlist_elem)
			{
				if(funcbit_test(cbelem->video_pktlist_elem->flags, MPXPLAY_PACKETLISTFLAG_CLEARBUF))
					*resetvideo = *resetfilter = true;
			}
			if(*resetvideo && funcbit_test(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_PRIMARYVIDEO))
				funcbit_enable(cbelem->flags, (DISPQT_FFMPGVIDEOCALLBACKFLAG_IGNORETIMESTAMP | DISPQT_FFMPGVIDEOCALLBACKFLAG_DISPLAYONEFRAME));
			*sleep_time = MPXPLAY_VIDEO_WORKER_THREAD_SLEEPTIME_SHORT;
		}
		else if(*video_ctx)
		{
			*sleep_time = MPXPLAY_VIDEO_WORKER_THREAD_SLEEPTIME_SHORT;
		}
	}
}

#define MPXPLAY_WORKDEC_SKIPFRAME_LEVEL_NONE    0
#define MPXPLAY_WORKDEC_SKIPFRAME_LEVEL_FILTER  1
#define MPXPLAY_WORKDEC_SKIPFRAME_LEVEL_ALL     2

static unsigned int video_worker_videodecoder_check_skip_frame(struct ffmpegvideo_callback_list_s *cbelem, unsigned int packet_flags)
{
	AVFrame *decoded_frame = cbelem->vid_frame_decoded;
	unsigned int skipframe_level = MPXPLAY_WORKDEC_SKIPFRAME_LEVEL_ALL;

	if(!decoded_frame)
	{
		return skipframe_level;
	}

	if( !funcbit_test(packet_flags, MPXPLAY_PACKETLISTFLAG_STILLPICT)
	 || (decoded_frame->width > cbelem->video_last_valid_frame_w)
	 || (decoded_frame->height > cbelem->video_last_valid_frame_h)
	){
		if(cbelem->video_delay_type == DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_SIGNIFICANT)
		{
			cbelem->video_delay_counter_workdec_filter = MPXPLAY_VIDEO_WORKDEC_FILTER_RESTORE_DELAY;
			skipframe_level = MPXPLAY_WORKDEC_SKIPFRAME_LEVEL_FILTER;
		}
		else if(cbelem->video_delay_counter_workdec_filter)
		{
			cbelem->video_delay_counter_workdec_filter--;
			skipframe_level = MPXPLAY_WORKDEC_SKIPFRAME_LEVEL_FILTER;
		}
		else
		{
			skipframe_level = MPXPLAY_WORKDEC_SKIPFRAME_LEVEL_NONE;
		}
	}
	cbelem->video_last_valid_frame_w = decoded_frame->width;
	cbelem->video_last_valid_frame_h = decoded_frame->height;

	return skipframe_level;
}

static int video_worker_videodecoder_filter_process(struct ffmpegvideo_callback_list_s *cbelem, AVFrame **vid_frame_filtered)
{
	int filter_retVal = -1;
	if(cbelem->video_frame_filter_workerprocess)
	{
		filter_retVal = cbelem->video_frame_filter_workerprocess->video_filter_apply(cbelem->vid_frame_decoded, vid_frame_filtered, AV_PIX_FMT_NONE);
		if(cbelem->vid_frame_decoded)
		{
			if(funcbit_test(filter_retVal, DISPQT_VIDEOFILTER_APPLYRETBIT_INPUTFRAME_PROCESSED))
			{
				av_frame_free(&cbelem->vid_frame_decoded);
			}
			else if(!funcbit_test(filter_retVal, DISPQT_VIDEOFILTER_APPLYRETBIT_OUTPUTFRAME_PRODUCED))
			{
				*vid_frame_filtered = cbelem->vid_frame_decoded;
				cbelem->vid_frame_decoded = NULL;
			}
		}
	}
	return filter_retVal;
}

static void video_worker_videodecoder_update_metadata_from_vframe(struct ffmpegvideo_callback_list_s *cbelem, AVFrame *vid_frame_decoded, unsigned int packet_flags)
{
	if(funcbit_test(packet_flags, MPXPLAY_PACKETLISTFLAG_STILLPICT) && vid_frame_decoded)
	{
		cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_SET_VIDFRAME_METADATA, (mpxp_ptrsize_t)vid_frame_decoded->metadata, 0, INFFMPG_MUTEX_COMMAND_TIMEOUT);
	}
}

static void video_worker_videodecoder_frameprocess_speedup_on_delay(struct ffmpegvideo_callback_list_s *cbelem, AVCodecContext *video_ctx)
{
	video_ctx->skip_loop_filter = AVDISCARD_DEFAULT;
	video_ctx->skip_frame = AVDISCARD_DEFAULT;
	switch(cbelem->video_delay_type)
	{
		case DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_SIGNIFICANT:
			video_ctx->skip_loop_filter = AVDISCARD_ALL;
			video_ctx->skip_frame = AVDISCARD_NONREF;
			//video_ctx->skip_frame = AVDISCARD_NONKEY; // FIXME: gives broken pictures
			break;
		case DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_LARGE:
			video_ctx->skip_frame = AVDISCARD_NONREF;//AVDISCARD_NONINTRA;
			break;
		case DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_MIDDLE:
			video_ctx->skip_frame = AVDISCARD_NONREF;//AVDISCARD_BIDIR;
			break;
		case DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_SMALL:
			break;
		default:
			video_ctx->skip_loop_filter = AVDISCARD_DEFAULT;
			break;
	}
}

#ifdef MPXPLAY_LINK_DXVA
/* retrieve data from GPU to CPU */
static int video_worker_videodecoder_gpu_retrieve_data(AVFrame *frame)
{
	AVFrame *tmp_frame = av_frame_alloc();
	int ret;

	ret = av_hwframe_transfer_data(tmp_frame, frame, 0);
	if (ret < 0)
		return ret;

	ret = av_frame_copy_props(tmp_frame, frame);
	if(ret < 0)
	{
		av_frame_unref(tmp_frame);
		return ret;
	}

	av_frame_unref(frame);
	av_frame_move_ref(frame, tmp_frame);

	return 0;
}
#endif

static void video_worker_videodecoder_frame_process(struct ffmpegvideo_callback_list_s *cbelem, AVCodecContext *video_ctx, bool *resetvideo, bool *resetfilter, unsigned int *sleep_time, unsigned int packet_flags)
{
	int retcode_video_receive = -1, retcode_video_send = -1, retry_videodec = 3, packet_used = 0, streamtype_index = MPXPLAY_STREAMTYPEINDEX_VIDEO;
	AVFrame *vid_frame_filtered = NULL;
	if(*resetfilter)
	{
		av_frame_free(&cbelem->vid_frame_decoded);
		if(cbelem->video_frame_filter_workerprocess)
			cbelem->video_frame_filter_workerprocess->video_filter_close();  // TODO: other way to clear?
		*resetfilter = false;
	}
	if(!video_ctx)
	{
		return;
	}

	do{
		unsigned int skipframe_level;
		if(!cbelem->vid_frame_decoded)
		{
			cbelem->vid_frame_decoded = av_frame_alloc();
			if(!cbelem->vid_frame_decoded)
				break;

			video_worker_videodecoder_frameprocess_speedup_on_delay(cbelem, video_ctx);

			retcode_video_receive = avcodec_receive_frame(video_ctx, cbelem->vid_frame_decoded);

			video_worker_videodecoder_update_metadata_from_vframe(cbelem, cbelem->vid_frame_decoded, packet_flags);
#if 0		// doesn't work
			if((retcode_video_receive < 0) || (cbelem->vid_frame_decoded->flags & AV_FRAME_FLAG_CORRUPT)) // || (vid_frame->decode_error_flags & (FF_DECODE_ERROR_INVALID_BITSTREAM | FF_DECODE_ERROR_MISSING_REFERENCE)))
			{
				if(retcode_video_receive != AVERROR(EAGAIN))
				{
					mpxplay_debugf(DISPQT_DEBUGOUT_WARNING, "avcodec_receive_frame ERROR %8.8X r:%d f:%d df:%d d:%8.8X ls:%d ", (unsigned int)cbelem,
							retcode_video_receive, cbelem->vid_frame_decoded->flags, cbelem->vid_frame_decoded->decode_error_flags, (unsigned int)cbelem->vid_frame_decoded->data[0], cbelem->vid_frame_decoded->linesize[0]);
				}
			}
#endif
#ifdef MPXPLAY_LINK_DXVA
			if( (retcode_video_receive == 0)
			 && ((cbelem->vid_frame_decoded->format == AV_PIX_FMT_DXVA2_VLD) || (cbelem->vid_frame_decoded->format == AV_PIX_FMT_D3D11)  || (cbelem->vid_frame_decoded->format == AV_PIX_FMT_D3D11VA_VLD))
			){
				retcode_video_receive = video_worker_videodecoder_gpu_retrieve_data(cbelem->vid_frame_decoded);
			}
#endif
			if((retcode_video_receive == 0) && (!cbelem->vid_frame_decoded->data[0] || (cbelem->vid_frame_decoded->linesize[0] <= 0) || (cbelem->vid_frame_decoded->flags & AV_FRAME_FLAG_DISCARD)))
				retcode_video_receive = -1;
			if(retcode_video_receive < 0)
				av_frame_free(&cbelem->vid_frame_decoded);
		}

		skipframe_level = video_worker_videodecoder_check_skip_frame(cbelem, packet_flags);
		if(skipframe_level == MPXPLAY_WORKDEC_SKIPFRAME_LEVEL_ALL)
		{
			av_frame_free(&cbelem->vid_frame_decoded);
		}
		else if(skipframe_level == MPXPLAY_WORKDEC_SKIPFRAME_LEVEL_FILTER)
		{
			vid_frame_filtered = cbelem->vid_frame_decoded;
			cbelem->vid_frame_decoded = NULL;
		}
		else if(video_worker_videodecoder_filter_process(cbelem, &vid_frame_filtered) < 0)
		{
			break;
		}

		if(vid_frame_filtered || packet_used || !cbelem->video_pktlist_elem)
		{
			break;
		}

		AVCodecContext *packet_vcodec_ctx = (AVCodecContext *)cbelem->video_pktlist_elem->codec_ctx;
		if(!packet_vcodec_ctx || !cbelem->video_pktlist_elem->frame_pkt || (cbelem->video_pktlist_elem->packet_type != MPXPLAY_PACKETTYPE_AVPACKET))
		{
			packet_used = 1;
			break;
		}
		if((packet_vcodec_ctx->codec_type != AVMEDIA_TYPE_VIDEO) || (packet_vcodec_ctx->codec_id == AV_CODEC_ID_NONE) || !avcodec_is_open(packet_vcodec_ctx) || !packet_vcodec_ctx->codec) // codec is not initialized (avcodec_open2 not called)
		{
			cbelem->videoctx_init_error_count++;
			if(cbelem->videoctx_init_error_count > INFFMPEG_VIDEO_INITERROR_LIMIT) // we drop the packet (containing the incorrect ctx) after a time
				packet_used = 1;
			break;
		}

		retcode_video_send = avcodec_send_packet(packet_vcodec_ctx, (AVPacket *)cbelem->video_pktlist_elem->frame_pkt);
		if(retcode_video_send == 0)
		{
			video_ctx = packet_vcodec_ctx;
			packet_used = 1;
		}
		else
		{
			mpxplay_debugf(DISPQT_DEBUG_ERROR, "avcodec_send_packet VIDEO failed ret:%d pctx:%8.8X vctx:%8.8X", retcode_video_send, packet_vcodec_ctx, video_ctx);
		}
	}while(--retry_videodec);

	if(packet_used || !retry_videodec) // clear packet if it's sent or seems to be unusefull
	{
		mpxplay_ffmpgdec_queuelistelem_clear(cbelem->video_pktlist_elem);
		cbelem->video_pktlist_elem = NULL;
		cbelem->videoctx_init_error_count = 0;
		*sleep_time = 0;
	}
	if(vid_frame_filtered)
	{
		if(vid_frame_filtered->width <= 0)
			vid_frame_filtered->width = video_ctx->width;
		if(vid_frame_filtered->height <= 0)
			vid_frame_filtered->height = video_ctx->height;
		if(vid_frame_filtered->format < 0)
			vid_frame_filtered->format = ((video_ctx->pix_fmt > AV_PIX_FMT_NONE)? video_ctx->pix_fmt : AV_PIX_FMT_YUV420P); // FIXME: default AV_PIX_FMT_YUV420P
		if(funcbit_test(packet_flags, MPXPLAY_PACKETLISTFLAG_STILLPICT)) // TODO: check: still pictures always use pixel aspect ratio override
		{
			vid_frame_filtered->sample_aspect_ratio.num = vid_frame_filtered->sample_aspect_ratio.den = 1;
		}
		else if((vid_frame_filtered->sample_aspect_ratio.num < video_ctx->sample_aspect_ratio.num) || (vid_frame_filtered->sample_aspect_ratio.den < video_ctx->sample_aspect_ratio.den))
		{
			vid_frame_filtered->sample_aspect_ratio.num = video_ctx->sample_aspect_ratio.num;
			vid_frame_filtered->sample_aspect_ratio.den = video_ctx->sample_aspect_ratio.den;
		}

		if(*resetvideo)
		{
			cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_SET_VIDEOTIMESTAMP, (mpxp_ptrsize_t)vid_frame_filtered, 0, INFFMPG_MUTEX_COMMAND_TIMEOUT);
			funcbit_enable(packet_flags, MPXPLAY_PACKETLISTFLAG_CLEARBUF);
			*resetvideo = false;
		}
		else if(cbelem->video_delay_type >= DISPQT_FFMPGVIDEOCALLBACK_VIDEODELAYTYPE_SMALL)
		{
			cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_SET_VIDEOTIMESTAMP, (mpxp_ptrsize_t)vid_frame_filtered, 0, INFFMPG_MUTEX_ONTIMER_TIMEOUT);
		}

		cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_GET_CORRECTEDVTIMESTAMP, (mpxp_ptrsize_t)vid_frame_filtered, (mpxp_ptrsize_t)&vid_frame_filtered->pts, INFFMPG_MUTEX_COMMAND_TIMEOUT);

		mpxplay_ffmpgdec_packetqueue_put(&cbelem->videoframe_queue, MPXPLAY_PACKETTYPE_AVFRAME_VIDEO, packet_flags, video_ctx, vid_frame_filtered);

		cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_SET_CODECCTX_CONSOLIDATE, (mpxp_ptrsize_t)video_ctx, streamtype_index, INFFMPG_MUTEX_COMMAND_TIMEOUT);
		*sleep_time = 0;
		mpxplay_debugf(DISPQT_DEBUGOUT_VTHREAD, "video_decoder_worker_thread decode DONE w:%d h:%d pts:%lld skip:%d", vid_frame_filtered->width, vid_frame_filtered->height, vid_frame_filtered->pts, video_ctx->skip_frame);
	}
}

static void video_worker_audiovisualization_process(struct ffmpegvideo_callback_list_s *cbelem, bool *resetvideo, unsigned int *sleep_time, unsigned int packet_flags)
{
	int streamtype_index = MPXPLAY_STREAMTYPEINDEX_VIDEO;
	bool audio_input_processed = true;
	if(!cbelem->video_pktlist_elem)
		cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_GET_PACKETLIST_ELEM, (mpxp_ptrsize_t)&cbelem->video_pktlist_elem, streamtype_index, INFFMPG_MUTEX_PACKETGET_TIMEOUT);
	if(cbelem->video_pktlist_elem)
	{
		AVFrame *audio_frame_decoded = (AVFrame *)cbelem->video_pktlist_elem->frame_pkt;
		if((cbelem->video_pktlist_elem->flags & MPXPLAY_PACKETLISTFLAG_CLEARBUF) || (cbelem->videoframe_queue.nb_packets > DISPQT_FFMPGVIDEO_MAX_DELAY_VISVIDEOFRAME))
		{
			funcbit_disable(cbelem->video_pktlist_elem->flags, MPXPLAY_PACKETLISTFLAG_CLEARBUF);
			*resetvideo = true;
		}
		//mpxplay_debugf(DISPQT_DEBUGOUT_VTHREAD, "audio visualization pe:%8.8X afd:%8.8X n:%d", (int)cbelem->video_pktlist_elem, (int)audio_frame_decoded, cbelem->videoframe_queue.nb_packets);
		if(audio_frame_decoded && (cbelem->video_pktlist_elem->packet_type == MPXPLAY_PACKETTYPE_AVFRAME_AUDIO) && (cbelem->videoframe_queue.nb_packets < DISPQT_FFMPGVIDEO_MAX_QUEUED_VISVIDEOFRAME) && cbelem->video_frame_filter_workerprocess)
		{
			AVFrame *vid_frame_visual = NULL;
			int retcode = cbelem->video_frame_filter_workerprocess->video_filter_apply(audio_frame_decoded, &vid_frame_visual, DISPQT_VIDEO_FILTERPROCESS_AVFORMAT);
			if(vid_frame_visual)
			{
				if(*resetvideo)
				{
					cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_SET_VIDEOTIMESTAMP, (mpxp_ptrsize_t)vid_frame_visual, 0, INFFMPG_MUTEX_COMMAND_TIMEOUT);
					cbelem->video_frame_filter_workerprocess->video_filter_close();  // TODO: other way to reset?
					funcbit_enable(packet_flags, MPXPLAY_PACKETLISTFLAG_CLEARBUF);
					funcbit_enable(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_IGNORETIMESTAMP);
					*resetvideo = false;
					mpxplay_debugf(DISPQT_DEBUGOUT_VTHREAD, "audio visualization RESET n:%d w:%d h:%d pts:%lld", cbelem->videoframe_queue.nb_packets, vid_frame_visual->width, vid_frame_visual->height, vid_frame_visual->pts);
				}
				// FIXME: this is a workaround for large audio blocks + visual plugins
				else
				{
#if 1
					mpxp_int64_t diff = vid_frame_visual->pts - cbelem->video_dec_last_timestamp;
					mpxplay_debugf(DISPQT_DEBUGOUT_VTHREAD, "vid_frame_visual pts:%lld DIFF:%lld", vid_frame_visual->pts, diff);
					if((diff > -(AV_TIME_BASE / 24)) && (diff < (AV_TIME_BASE / 50))) // visual plugin shall send with 25 fps
					{
						vid_frame_visual->pts = cbelem->video_dec_last_timestamp + (AV_TIME_BASE / 26);
					}
#endif
					if(!funcbit_test(retcode, DISPQT_VIDEOFILTER_APPLYRETBIT_INPUTFRAME_PROCESSED))
					{
						audio_input_processed = false;
					}
				}
				cbelem->video_dec_last_timestamp = vid_frame_visual->pts;
				//funcbit_enable(packet_flags, MPXPLAY_PACKETLISTFLAG_IGNORE_TS);
				//funcbit_enable(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_IGNORETIMESTAMP); // for testing only
				funcbit_enable(vid_frame_visual->flags, (DISPQT_AVFRAME_FLAG_NOSKIPFRAME | DISPQT_AVFRAME_FLAG_AR_WINDOW));
				if(mpxplay_ffmpgdec_packetqueue_put(&cbelem->videoframe_queue, MPXPLAY_PACKETTYPE_AVFRAME_VIDEO, packet_flags, NULL, vid_frame_visual) < 0)
				{
					av_frame_free(&vid_frame_visual);
				}
				//mpxplay_debugf(DISPQT_DEBUGOUT_VTHREAD, "mpxplay_ffmpgdec_packetqueue_put DONE n:%d w:%d h:%d pts:%lld", cbelem->videoframe_queue.nb_packets, vid_frame_visual->width, vid_frame_visual->height, vid_frame_visual->pts);
				*sleep_time = MPXPLAY_VIDEO_WORKER_THREAD_SLEEPTIME_SHORT;
			}
			else
			{
				*sleep_time = 0;
			}
		}

		if(audio_input_processed)
		{
			mpxplay_ffmpgdec_queuelistelem_clear(cbelem->video_pktlist_elem);
			cbelem->video_pktlist_elem = NULL;
		}
	}
}

#if defined(PDS_THREADS_POSIX_THREAD)
void *mpxplay_dispqt_video_decoder_worker_thread(void *decoder_data)
#else
unsigned int mpxplay_dispqt_video_decoder_worker_thread(void *decoder_data)
#endif
{
	struct ffmpegvideo_callback_list_s *cbelem = (struct ffmpegvideo_callback_list_s *)decoder_data;
	int streamtype_index = MPXPLAY_STREAMTYPEINDEX_VIDEO, flags = 0, cberror_count = 0;
	bool resetvideo = false, resetfilter = false;
	unsigned int packet_flags, sleep_time;
	AVCodecContext *video_ctx;

	do{
		video_ctx = NULL;
		sleep_time = MPXPLAY_VIDEO_WORKER_THREAD_SLEEPTIME_DEFAULT;
		if(!cbelem || !cbelem->infile_callback)
		{
			if(++cberror_count < MPXPLAY_VIDEO_WORKER_THREAD_CRITICAL_ERROR_COUNT)
			{
				pds_threads_sleep(sleep_time);
				continue;
			}
			mpxplay_debugf(DISPQT_DEBUG_ERROR, "video_decoder_worker_thread ERROR cbelem");
			break;
		}
		cberror_count = 0;
		packet_flags = 0;
		video_worker_videodecoder_context_process(cbelem, &video_ctx, &resetvideo, &resetfilter, &sleep_time, &packet_flags);

		if(funcbit_test(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_PRIMARYVIDEO))
		{
			if((mpxplay_config_video_audiovisualization_type == DISPQT_AUDIOVISUALIZATIONTYPE_NONE) || video_ctx) // TODO: extra switch to show audio visualization for video files too
			{
				if((cbelem->videoframe_queue.nb_packets < DISPQT_FFMPGVIDEO_MAX_QUEUED_VIDEOFRAME) || resetvideo)
				{
					video_worker_videodecoder_frame_process(cbelem, video_ctx, &resetvideo, &resetfilter, &sleep_time, packet_flags);
				}
			}
			else if(funcbit_test(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_PRIMARYVIDEO))
			{
				video_worker_audiovisualization_process(cbelem, &resetvideo, &sleep_time, packet_flags); // audio visualization
			}
			if(resetvideo)
			{
				mpxplay_ffmpgdec_packetqueue_put(&cbelem->videoframe_queue, MPXPLAY_PACKETTYPE_AVFRAME_VIDEO, MPXPLAY_PACKETLISTFLAG_CLEARBUF, NULL, NULL);
			}
		}
		else
		{ // TODO: currently we simply drop non primary video packets (later can be video crossfade or ...)
			if(cbelem->video_pktlist_elem)
			{
				mpxplay_ffmpgdec_queuelistelem_clear(cbelem->video_pktlist_elem);
				cbelem->video_pktlist_elem = NULL;
			}
			resetvideo = false;
			mpxplay_debugf(DISPQT_DEBUGOUT_VTHREAD, "video_decoder_worker_thread DROP END");
		}
		if(sleep_time > 0)
		{
			pds_threads_sleep(sleep_time);
		}
	}while(!funcbit_test(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_EXITEVENT));

	if(video_ctx)
	{
		if(avcodec_is_open(video_ctx) && video_ctx->codec)
			avcodec_flush_buffers(video_ctx);
		cbelem->infile_callback(cbelem->infile_passdata, MPXPLAY_INFILE_CBKCTRL_SET_CODECCTX_CLEAR, streamtype_index, flags, INFFMPG_MUTEX_COMMAND_TIMEOUT);
	}
	mpxplay_debugf(DISPQT_DEBUGOUT_VFCC, "video_decoder_worker_thread END");
	funcbit_enable(cbelem->flags, DISPQT_FFMPGVIDEOCALLBACKFLAG_THEXITDONE);
	pds_threads_sleep(DISPQT_VIDEO_MUTEX_TIMEOUT);  // required for pds_threads_thread_close -> TerminateThread
	return 0;
}

#endif // MPXPLAY_LINK_ORIGINAL_FFMPEG
